/*
 * Copyright (c) 2021-2023 Datalayer, Inc.
 *
 * MIT License
 */
import { Signal } from '@lumino/signaling';
import { OutputAreaModel } from '@jupyterlab/outputarea';
import { outputsAsString } from '../../utils/Utils';
export class KernelExecutor {
    _kernelConnection;
    _outputs;
    _outputsChanged = new Signal(this);
    _model;
    _modelChanged = new Signal(this);
    _executeReplyReceived = new Signal(this);
    _future;
    _shellMessageHooks;
    _executed;
    _executedResolve;
    constructor(kernelConnection) {
        this._kernelConnection = kernelConnection;
        this._outputs = [];
        this._model = new OutputAreaModel();
        this._executed = new Promise((resolve, _) => {
            this._executedResolve = resolve;
        });
    }
    execute(code, iopubMessageHooks = [], shellMessageHooks = []) {
        const future = this._kernelConnection.requestExecute({
            code,
        });
        iopubMessageHooks.forEach(hook => future.registerMessageHook(hook));
        this._shellMessageHooks = shellMessageHooks;
        this.future = future;
        return {
            future,
            executor: this,
        };
    }
    _onIOPub = (message) => {
        if (this._future?.msg.header.msg_id !== message.parent_header.msg_id) {
            return;
        }
        const messageType = message.header.msg_type;
        const output = { ...message.content, output_type: messageType };
        switch (messageType) {
            case 'execute_result':
                this._outputs.push(message.content);
                this._outputsChanged.emit(this._outputs);
                this._model.add(output);
                this._modelChanged.emit(this._model);
                break;
            case 'stream':
                this._outputs.push(message.content);
                this._outputsChanged.emit(this._outputs);
                this._model.add(output);
                this._modelChanged.emit(this._model);
                break;
            case 'display_data':
                this._outputs.push(message.content);
                this._outputsChanged.emit(this._outputs);
                this._model.add(output);
                this._modelChanged.emit(this._model);
                break;
            case 'update_display_data':
                this._outputs.push(message.content);
                this._outputsChanged.emit(this._outputs);
                this._model.add(output);
                this._modelChanged.emit(this._model);
                break;
            case 'status':
                {
                    // execution_state: 'busy' 'starting' 'terminating' 'restarting' 'initializing' 'connecting' 'disconnected' 'dead' 'unknown' 'idle'
                    const executionState = message.content.execution_state;
                    executionState;
                }
                break;
            default:
                break;
        }
    };
    _onReply = (message) => {
        if (this._future?.msg.header.msg_id !== message.parent_header.msg_id) {
            return;
        }
        const messageType = message.header.msg_type;
        this._shellMessageHooks.forEach(hook => hook(message));
        switch (messageType) {
            case 'execute_reply':
                {
                    const output = {
                        output_type: 'display_data',
                        data: message.content.data,
                        metadata: {},
                    };
                    this._outputs.push(message.content);
                    this._outputsChanged.emit(this._outputs);
                    this._model.add(output);
                    this.executeReplyReceived.emit(this._model);
                    this._executedResolve(this._model);
                }
                break;
            default:
                break;
        }
    };
    registerIOPubMessageHook = (msg) => {
        this._future?.registerMessageHook(msg);
    };
    get executed() {
        return this._executed;
    }
    get future() {
        return this._future;
    }
    set future(value) {
        this._future = value;
        if (!value) {
            return;
        }
        value.onIOPub = this._onIOPub;
        value.onReply = this._onReply;
    }
    get result() {
        return this.executed.then(model => {
            return outputsAsString(model.toJSON());
        });
    }
    get outputs() {
        return this._outputs;
    }
    get outputsChanged() {
        return this._outputsChanged;
    }
    get model() {
        return this._model;
    }
    get modelChanged() {
        return this._modelChanged;
    }
    get executeReplyReceived() {
        return this._executeReplyReceived;
    }
    get done() {
        return this._future?.done;
    }
}
export default KernelExecutor;
//# sourceMappingURL=KernelExecutor.js.map